# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: buildgrid/v2/introspection.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2 as build_dot_bazel_dot_remote_dot_execution_dot_v2_dot_remote__execution__pb2
from buildgrid._protos.google.devtools.remoteworkers.v1test2 import bots_pb2 as google_dot_devtools_dot_remoteworkers_dot_v1test2_dot_bots__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n buildgrid/v2/introspection.proto\x12\x0c\x62uildgrid.v2\x1a\x36\x62uild/bazel/remote/execution/v2/remote_execution.proto\x1a\x30google/devtools/remoteworkers/v1test2/bots.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"a\n\x12ListWorkersRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x0c\n\x04page\x18\x03 \x01(\x05\x12\x13\n\x0bworker_name\x18\x04 \x01(\t\")\n\x0fWorkerOperation\x12\x16\n\x0eoperation_name\x18\x01 \x01(\t\"\xd9\x02\n\x10RegisteredWorker\x12\x14\n\x0csession_name\x18\x01 \x01(\t\x12\x13\n\x0bworker_name\x18\x02 \x01(\t\x12\x30\n\x0clast_updated\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x44\n\nbot_status\x18\x04 \x01(\x0e\x32\x30.google.devtools.remoteworkers.v1test2.BotStatus\x12/\n\x0b\x65xpiry_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\noperations\x18\x06 \x03(\x0b\x32\x1d.buildgrid.v2.WorkerOperation\x12>\n\raction_digest\x18\x07 \x01(\x0b\x32\'.build.bazel.remote.execution.v2.Digest\"v\n\x13ListWorkersResponse\x12\x0c\n\x04page\x18\x01 \x01(\x05\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\r\n\x05total\x18\x03 \x01(\x05\x12/\n\x07workers\x18\x04 \x03(\x0b\x32\x1e.buildgrid.v2.RegisteredWorker\"3\n\x1aGetOperationFiltersRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\"t\n\x0fOperationFilter\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x13\n\x0b\x63omparators\x18\x05 \x03(\t\x12\x0e\n\x06values\x18\x06 \x03(\t\"B\n\x10OperationFilters\x12.\n\x07\x66ilters\x18\x01 \x03(\x0b\x32\x1d.buildgrid.v2.OperationFilter2\xc8\x01\n\rIntrospection\x12T\n\x0bListWorkers\x12 .buildgrid.v2.ListWorkersRequest\x1a!.buildgrid.v2.ListWorkersResponse\"\x00\x12\x61\n\x13GetOperationFilters\x12(.buildgrid.v2.GetOperationFiltersRequest\x1a\x1e.buildgrid.v2.OperationFilters\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.buildgrid.v2.introspection_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_LISTWORKERSREQUEST']._serialized_start=189
  _globals['_LISTWORKERSREQUEST']._serialized_end=286
  _globals['_WORKEROPERATION']._serialized_start=288
  _globals['_WORKEROPERATION']._serialized_end=329
  _globals['_REGISTEREDWORKER']._serialized_start=332
  _globals['_REGISTEREDWORKER']._serialized_end=677
  _globals['_LISTWORKERSRESPONSE']._serialized_start=679
  _globals['_LISTWORKERSRESPONSE']._serialized_end=797
  _globals['_GETOPERATIONFILTERSREQUEST']._serialized_start=799
  _globals['_GETOPERATIONFILTERSREQUEST']._serialized_end=850
  _globals['_OPERATIONFILTER']._serialized_start=852
  _globals['_OPERATIONFILTER']._serialized_end=968
  _globals['_OPERATIONFILTERS']._serialized_start=970
  _globals['_OPERATIONFILTERS']._serialized_end=1036
  _globals['_INTROSPECTION']._serialized_start=1039
  _globals['_INTROSPECTION']._serialized_end=1239
# @@protoc_insertion_point(module_scope)
