# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: build/buildbox/execution_stats.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$build/buildbox/execution_stats.proto\x12\x0e\x62uild.buildbox\x1a\x1egoogle/protobuf/duration.proto\"\xc5\x02\n\x13\x45xecutionStatistics\x12P\n\x0e\x63ommand_rusage\x18\x01 \x01(\x0b\x32\x38.build.buildbox.ExecutionStatistics.ProcessResourceUsage\x1a\xdb\x01\n\x14ProcessResourceUsage\x12(\n\x05utime\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\x12(\n\x05stime\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x0e\n\x06maxrss\x18\x03 \x01(\x03\x12\x0e\n\x06minflt\x18\x04 \x01(\x03\x12\x0e\n\x06majflt\x18\x05 \x01(\x03\x12\x0f\n\x07inblock\x18\x06 \x01(\x03\x12\x0f\n\x07oublock\x18\x07 \x01(\x03\x12\r\n\x05nvcsw\x18\x08 \x01(\x03\x12\x0e\n\x06nivcsw\x18\t \x01(\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.build.buildbox.execution_stats_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EXECUTIONSTATISTICS']._serialized_start=89
  _globals['_EXECUTIONSTATISTICS']._serialized_end=414
  _globals['_EXECUTIONSTATISTICS_PROCESSRESOURCEUSAGE']._serialized_start=195
  _globals['_EXECUTIONSTATISTICS_PROCESSRESOURCEUSAGE']._serialized_end=414
# @@protoc_insertion_point(module_scope)
