# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/devtools/build/v1/build_events.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from buildgrid._protos.google.devtools.build.v1 import build_status_pb2 as google_dot_devtools_dot_build_dot_v1_dot_build__status__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+google/devtools/build/v1/build_events.proto\x12\x18google.devtools.build.v1\x1a+google/devtools/build/v1/build_status.proto\x1a\x19google/protobuf/any.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xd0\x0b\n\nBuildEvent\x12.\n\nevent_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x63\n\x1ainvocation_attempt_started\x18\x33 \x01(\x0b\x32=.google.devtools.build.v1.BuildEvent.InvocationAttemptStartedH\x00\x12\x65\n\x1binvocation_attempt_finished\x18\x34 \x01(\x0b\x32>.google.devtools.build.v1.BuildEvent.InvocationAttemptFinishedH\x00\x12L\n\x0e\x62uild_enqueued\x18\x35 \x01(\x0b\x32\x32.google.devtools.build.v1.BuildEvent.BuildEnqueuedH\x00\x12L\n\x0e\x62uild_finished\x18\x37 \x01(\x0b\x32\x32.google.devtools.build.v1.BuildEvent.BuildFinishedH\x00\x12L\n\x0e\x63onsole_output\x18\x38 \x01(\x0b\x32\x32.google.devtools.build.v1.BuildEvent.ConsoleOutputH\x00\x12\x66\n\x19\x63omponent_stream_finished\x18; \x01(\x0b\x32\x41.google.devtools.build.v1.BuildEvent.BuildComponentStreamFinishedH\x00\x12+\n\x0b\x62\x61zel_event\x18< \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x12\x35\n\x15\x62uild_execution_event\x18= \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x12\x32\n\x12source_fetch_event\x18> \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x1aY\n\x18InvocationAttemptStarted\x12\x16\n\x0e\x61ttempt_number\x18\x01 \x01(\x03\x12%\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\x84\x01\n\x19InvocationAttemptFinished\x12@\n\x11invocation_status\x18\x03 \x01(\x0b\x32%.google.devtools.build.v1.BuildStatus\x12%\n\x07\x64\x65tails\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\x36\n\rBuildEnqueued\x12%\n\x07\x64\x65tails\x18\x01 \x01(\x0b\x32\x14.google.protobuf.Any\x1am\n\rBuildFinished\x12\x35\n\x06status\x18\x01 \x01(\x0b\x32%.google.devtools.build.v1.BuildStatus\x12%\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\x86\x01\n\rConsoleOutput\x12;\n\x04type\x18\x01 \x01(\x0e\x32-.google.devtools.build.v1.ConsoleOutputStream\x12\x15\n\x0btext_output\x18\x02 \x01(\tH\x00\x12\x17\n\rbinary_output\x18\x03 \x01(\x0cH\x00\x42\x08\n\x06output\x1a\xc0\x01\n\x1c\x42uildComponentStreamFinished\x12Z\n\x04type\x18\x01 \x01(\x0e\x32L.google.devtools.build.v1.BuildEvent.BuildComponentStreamFinished.FinishType\"D\n\nFinishType\x12\x1b\n\x17\x46INISH_TYPE_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x46INISHED\x10\x01\x12\x0b\n\x07\x45XPIRED\x10\x02\x42\x07\n\x05\x65vent\"\xc8\x01\n\x08StreamId\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\x12\x15\n\rinvocation_id\x18\x06 \x01(\t\x12\x44\n\tcomponent\x18\x03 \x01(\x0e\x32\x31.google.devtools.build.v1.StreamId.BuildComponent\"M\n\x0e\x42uildComponent\x12\x15\n\x11UNKNOWN_COMPONENT\x10\x00\x12\x0e\n\nCONTROLLER\x10\x01\x12\n\n\x06WORKER\x10\x02\x12\x08\n\x04TOOL\x10\x03*:\n\x13\x43onsoleOutputStream\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06STDOUT\x10\x01\x12\n\n\x06STDERR\x10\x02\x42\x8b\x01\n\x1c\x63om.google.devtools.build.v1B\x0f\x42uildEventProtoP\x01Z=google.golang.org/genproto/googleapis/devtools/build/v1;build\xf8\x01\x01\xca\x02\x15Google\\Cloud\\Build\\V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.google.devtools.build.v1.build_events_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.devtools.build.v1B\017BuildEventProtoP\001Z=google.golang.org/genproto/googleapis/devtools/build/v1;build\370\001\001\312\002\025Google\\Cloud\\Build\\V1'
  _globals['_CONSOLEOUTPUTSTREAM']._serialized_start=1872
  _globals['_CONSOLEOUTPUTSTREAM']._serialized_end=1930
  _globals['_BUILDEVENT']._serialized_start=179
  _globals['_BUILDEVENT']._serialized_end=1667
  _globals['_BUILDEVENT_INVOCATIONATTEMPTSTARTED']._serialized_start=935
  _globals['_BUILDEVENT_INVOCATIONATTEMPTSTARTED']._serialized_end=1024
  _globals['_BUILDEVENT_INVOCATIONATTEMPTFINISHED']._serialized_start=1027
  _globals['_BUILDEVENT_INVOCATIONATTEMPTFINISHED']._serialized_end=1159
  _globals['_BUILDEVENT_BUILDENQUEUED']._serialized_start=1161
  _globals['_BUILDEVENT_BUILDENQUEUED']._serialized_end=1215
  _globals['_BUILDEVENT_BUILDFINISHED']._serialized_start=1217
  _globals['_BUILDEVENT_BUILDFINISHED']._serialized_end=1326
  _globals['_BUILDEVENT_CONSOLEOUTPUT']._serialized_start=1329
  _globals['_BUILDEVENT_CONSOLEOUTPUT']._serialized_end=1463
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED']._serialized_start=1466
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED']._serialized_end=1658
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED_FINISHTYPE']._serialized_start=1590
  _globals['_BUILDEVENT_BUILDCOMPONENTSTREAMFINISHED_FINISHTYPE']._serialized_end=1658
  _globals['_STREAMID']._serialized_start=1670
  _globals['_STREAMID']._serialized_end=1870
  _globals['_STREAMID_BUILDCOMPONENT']._serialized_start=1793
  _globals['_STREAMID_BUILDCOMPONENT']._serialized_end=1870
# @@protoc_insertion_point(module_scope)
