"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright (C) 2021 Bloomberg LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 <http://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import buildgrid._protos.google.devtools.build.v1.build_events_pb2
import buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class QueryEventStreamsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUILD_ID_PATTERN_FIELD_NUMBER: builtins.int
    build_id_pattern: builtins.str
    def __init__(
        self,
        *,
        build_id_pattern: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["build_id_pattern", b"build_id_pattern"]) -> None: ...

global___QueryEventStreamsRequest = QueryEventStreamsRequest

@typing_extensions.final
class QueryEventStreamsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class BuildEventStream(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        STREAM_ID_FIELD_NUMBER: builtins.int
        EVENTS_FIELD_NUMBER: builtins.int
        @property
        def stream_id(self) -> buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId: ...
        @property
        def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2.OrderedBuildEvent]: ...
        def __init__(
            self,
            *,
            stream_id: buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId | None = ...,
            events: collections.abc.Iterable[buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2.OrderedBuildEvent] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["stream_id", b"stream_id"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["events", b"events", "stream_id", b"stream_id"]) -> None: ...

    STREAMS_FIELD_NUMBER: builtins.int
    @property
    def streams(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___QueryEventStreamsResponse.BuildEventStream]: ...
    def __init__(
        self,
        *,
        streams: collections.abc.Iterable[global___QueryEventStreamsResponse.BuildEventStream] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["streams", b"streams"]) -> None: ...

global___QueryEventStreamsResponse = QueryEventStreamsResponse
