# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add property label

Revision ID: 23f48374c8b6
Revises: fe3a61461feb
Create Date: 2024-09-05 11:49:28.973640

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "23f48374c8b6"
down_revision = "fe3a61461feb"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("jobs", sa.Column("property_label", sa.String(), server_default="unknown", nullable=False))
    op.drop_index("ix_jobs_stage", table_name="jobs")
    op.create_index("ix_jobs_stage_property_label", "jobs", ["stage", "property_label"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_jobs_stage_property_label", table_name="jobs")
    op.create_index("ix_jobs_stage", "jobs", ["stage"], unique=False)
    op.drop_column("jobs", "property_label")
    # ### end Alembic commands ###
